/*
 * This file is part of the Disco Deterministic Network Calculator v2.0.1 "Hydra".
 *
 * Copyright (C) 2011 - 2014 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.demos;

import java.util.Iterator;

import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.nc.TotalFlowAnalysis;
import unikl.disco.nc.SeparateFlowAnalysis;
import unikl.disco.nc.PmooAnalysis;

import unikl.disco.network.Link;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Server;

/**
 * 
 * @author Steffen Bondorf
 */
public class Demo2
{
	public Demo2() {}

	@SuppressWarnings("unused")
	public void run()
	{
		ServiceCurve service_curve = ServiceCurve.createRateLatency( 10.0e6, 0.01 );
		ServiceCurve max_service_curve = ServiceCurve.createRateLatency( 100.0e6, 0.001 );
		
		Network network = new Network();
		
		try {
			Server s0 = network.addServer( service_curve, max_service_curve );
			s0.setUseGamma( false );
			s0.setUseExtraGamma( false );
			
			Server s1 = network.addServer( service_curve, max_service_curve );
			s1.setUseGamma( false );
			s1.setUseExtraGamma( false );
		
			Link l_0_1 = network.addLink( s0, s1 );
	
			ArrivalCurve arrival_curve = ArrivalCurve.createTokenBucket( 0.1e6, 0.1 * 0.1e6 );
			
			Flow f0 = network.addFlow( arrival_curve, s1 );

			Flow f1 = network.addFlow( arrival_curve, s0, s1 );
			
			Iterator<Flow> flow_iter = network.getFlows().iterator();
			while ( flow_iter.hasNext() ) {
				Flow flow_of_interest = flow_iter.next();
			
				System.out.println( "Flow of interest : " + flow_of_interest.toString() );
				System.out.println();

//				Analyze the network
//				TFA
				System.out.println( "--- Total Flow Analysis ---" );				
				TotalFlowAnalysis tfa = new TotalFlowAnalysis( network );
				
				try {
					tfa.performEnd2EndAnalysis( flow_of_interest );
					System.out.println( "delay bound     : " + tfa.getDelayBound() );
					System.out.println( "backlog bound   : " + tfa.getBacklogBound() );
					System.out.println( "     per server : " + tfa.getServerBacklogBoundMap().toString() );
				} catch (Exception e) {
					System.out.println( "TFA analysis failed" );
					System.out.println( e.toString() );
				}
				
				System.out.println();

//				SFA
				System.out.println( "--- Separated Flow Analysis ---" );
				SeparateFlowAnalysis sfa = new SeparateFlowAnalysis( network );
				
				try {
					sfa.performEnd2EndAnalysis( flow_of_interest );
					System.out.println( "e2e SFA SC      : " + sfa.getLeftOverServiceCurves() );
					System.out.println( "     per server : " + sfa.getServerLeftOverBetasMap().toString() );
					System.out.println( "delay bound     : " + sfa.getDelayBound() );
					System.out.println( "backlog bound   : " + sfa.getBacklogBound() );
				} catch (Exception e) {
					System.out.println( "SFA analysis failed" );
					System.out.println( e.toString() );
				}
				
				System.out.println();
				
//				PMOO
				System.out.println( "--- PMOO Analysis ---" );
				PmooAnalysis pmoo = new PmooAnalysis( network );
				
				try {
					pmoo.performEnd2EndAnalysis( flow_of_interest );
					System.out.println( "e2e PMOO SFA SC : " + pmoo.getLeftOverServiceCurves() );
					System.out.println( "delay bound     : " + pmoo.getDelayBound() );
					System.out.println( "backlog bound   : " + pmoo.getBacklogBound() );
				} catch (Exception e) {
					System.out.println( "PMOO SFA analysis failed" );
					System.out.println( e.toString() );
				}

				System.out.println();
				System.out.println();
			}
		} catch (Exception e) {
			System.out.println( e.toString() );
		}
	}

	public static void main( String[] args )
	{
		Demo2 demo = new Demo2();
		demo.run();
	}
}
