/*
 * This file is part of the Disco Deterministic Network Calculator v2.0.1 "Hydra".
 *
 * Copyright (C) 2005 - 2007 Frank A. Zdarsky
 * Copyright (C) 2008 - 2010 Andreas Kiefer
 * Copyright (C) 2011 - 2014 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.nc;

import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.Stack;

import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.misc.SetUtils;
import unikl.disco.nc.Configuration.ArrivalBoundMethods;
import unikl.disco.nc.Configuration.MuxDiscipline;
import unikl.disco.nc.Configuration.GammaFlag;
import unikl.disco.network.Flow;
import unikl.disco.network.Link;
import unikl.disco.network.Network;
import unikl.disco.network.Server;
import unikl.disco.minplus.Convolution;
import unikl.disco.minplus.Deconvolution;

/**
 * 
 * @author Frank A. Zdarsky
 * @author Andreas Kiefer
 * @author Steffen Bondorf
 *
 */
@Deprecated // This legacy code will not be maintained any longer. See WONTFIX annotations for details.
public class PbooArrivalBound_Output_PerHop extends ArrivalBound {
	Network network;
	Flow flow_of_interest;
	
	Stack<OutputCacheEntry> stack;
	OutputBoundCache internal_cache; 
	 
	@SuppressWarnings("unused")
	private PbooArrivalBound_Output_PerHop() {}

	public PbooArrivalBound_Output_PerHop( Network network, Flow flow_of_interest ) {
		this.network = network;
		this.flow_of_interest = flow_of_interest;
	}

	protected ArrivalCurve computeArrivalBound( Server server ) throws Exception {
		return computeArrivalBound( server, network.getFlows( server ) );
	}

	protected ArrivalCurve computeArrivalBound( Link link ) throws Exception {
		return computeArrivalBound( link.getDest(), network.getFlows( link ) );
	}
	
	protected ArrivalCurve computeArrivalBound( Link link, Set<Flow> f_xfcaller ) throws Exception {
		Set<Flow> flows_on_l_to_bound = SetUtils.getIntersection( f_xfcaller, network.getFlows( link ) );
		return computeArrivalBound( link.getDest(), flows_on_l_to_bound );
	}
	
	/**
	 * Returns the arrival bound for a set of <code>f_xfcaller</code> at
	 * the given <code>server</code>.
	 * 
	 * It is possible to change between iterative and recursive computation by
	 * calling <code>setUseRecursiveOutputBoundComputation()</code>.
	 * 
	 * @param server
	 *            The server for which the arrival bound is computed
	 * @param f_xfcaller
	 *            Set of flows for which the arrival bound is computed
	 * @return A <code>Curve</code> instance containing the output bound
	 * @throws OutOfMemoryError
	 */
	protected ArrivalCurve computeArrivalBound( Server server, Set<Flow> f_xfcaller ) throws Exception
	{
		ArrivalCurve alpha_xfcaller = ArrivalCurve.createNullArrival();
		
		// Get flows to bound that are actually present at server
		Set<Flow> f_xfcaller_server = SetUtils.getIntersection( f_xfcaller, network.getFlows( server ) );
		f_xfcaller_server.remove( flow_of_interest );
		if ( f_xfcaller_server.size() == 0 )
		{
			return alpha_xfcaller;
		}
		
		// Get cross-traffic originating in server
		Set<Flow> f_xfcaller_sourceflows_server = SetUtils.getIntersection( f_xfcaller_server, network.getSourceFlows( server ) );
		alpha_xfcaller = network.getSourceFlowArrivalCurve( server, f_xfcaller_sourceflows_server );
		
		// Get cross-traffic from each predecessor
		ArrivalCurve arrival_bound_link = ArrivalCurve.createNullArrival();
		Set<Flow> f_xfcaller_server_bounded = f_xfcaller_sourceflows_server;
		for ( Link in_l : network.getInLinks( server ) )
		{
			Set<Flow> f_xfcaller_link = SetUtils.getIntersection( network.getFlows( in_l ), f_xfcaller_server );
			
			if( f_xfcaller_link.isEmpty() ) { // Do not check links without flows of interest
				continue;
			}
			
			if ( Configuration.useRecursiveOutputBasedArrivalBoundComputation() )
			{
				arrival_bound_link = computeRecursiveOutputBound( in_l.getSource(), f_xfcaller_link ); // Semantics: OutputBound of link's source
			}
			else
			{
				arrival_bound_link = computeIterativeOutputBound( in_l.getSource(), f_xfcaller_link ); // Semantics: OutputBound of link's source
			}
			alpha_xfcaller = ArrivalCurve.add( alpha_xfcaller, arrival_bound_link );
			
			f_xfcaller_server_bounded.addAll( f_xfcaller_link ); // Stop as soon as all flows are bounded
			if ( f_xfcaller_server_bounded.containsAll( f_xfcaller_server ) ) {
				break;
			}
		}
		
		return alpha_xfcaller;
	}
	
	/**
	 * Returns the arrival bound for a set of <code>f_xfcaller</code> from
	 * the given <code>server</code> by recursion.<br/>
	 * 
	 * @param server
	 *            The server at which the output bound is computed
	 * @param f_xfcaller
	 *            Set of flows for which the output bound is computed
	 * @return A <code>Curve</code> instance containing the output bound
	 * 
	 * @throws OutOfMemoryError
	 */
	public ArrivalCurve computeRecursiveOutputBound( Server server, Set<Flow> f_xfcaller ) throws OutOfMemoryError
	{
		ArrivalCurve alpha_xfcaller = ArrivalCurve.createNullArrival();
		f_xfcaller.remove( flow_of_interest );

		if ( f_xfcaller.size() == 0 )
		{
			return alpha_xfcaller;
		}
		
		// Get flows originating in this server and sort by 
		// flows that are cross traffic at the next server (f_xfcaller_server, i.e, cross traffic of the caller)
		// flow that are cross traffic to them (f_xxfcaller_server, i.e., cross traffic to the cross traffic of the caller)
		Set<Flow> f_server = network.getFlows( server );
		Set<Flow> f_xfcaller_server = SetUtils.getIntersection( f_server, f_xfcaller );
		
		Set<Flow> f_xxfcaller_server = SetUtils.getDifference( f_server, f_xfcaller_server );
		f_xxfcaller_server.remove( flow_of_interest );

		ArrivalCurve alpha_xfcaller_link_sum = ArrivalCurve.createNullArrival();
		ArrivalCurve alpha_xxfcaller_link_sum = ArrivalCurve.createNullArrival();
		
		Set<Flow> f_link;
		Set<Flow> f_xfcaller_link;
		Set<Flow> f_xxfcaller_link;
		
		for ( Link in_l : network.getInLinks( server ) ) {
			f_link = network.getFlows( in_l );
			f_link.remove( flow_of_interest ); // Flow of interest has the lowest priority
			
			f_xfcaller_link = SetUtils.getIntersection( f_link, f_xfcaller_server );
			f_xxfcaller_link = SetUtils.getDifference( f_link, f_xfcaller_link );

			// FIXME WONTFIX: Handling of sets of arrival curves returned by ArrivalBound.computeArrivalBound( ... ).
			// Therefore, it is highly recommended not to combine this arrival bounding method with any other.  
			try {
				alpha_xfcaller_link_sum = ArrivalCurve.add( alpha_xfcaller_link_sum, ArrivalBound.computeArrivalBound( network, server, f_xfcaller_link, flow_of_interest )
						.iterator().next() );  
				alpha_xxfcaller_link_sum = ArrivalCurve.add( alpha_xxfcaller_link_sum, ArrivalBound.computeArrivalBound( network, server, f_xxfcaller_link, flow_of_interest )
						.iterator().next() );
			} catch (Exception e) {}
		}

		ArrivalCurve alpha_xfcaller_server = network.getSourceFlowArrivalCurve( server, f_xfcaller_server );
		ArrivalCurve alpha_xxfcaller_server = network.getSourceFlowArrivalCurve( server, f_xxfcaller_server );

		alpha_xfcaller_server = ArrivalCurve.add( alpha_xfcaller_server, alpha_xfcaller_link_sum );
		alpha_xxfcaller_server = ArrivalCurve.add( alpha_xxfcaller_server, alpha_xxfcaller_link_sum );
		
		alpha_xfcaller_server.beautify();
		alpha_xxfcaller_server.beautify();

		ServiceCurve beta_loxfcaller_server = server.getServiceCurve();
		ServiceCurve gamma = server.getMaxServiceCurve();
		if( Configuration.multiplexingDiscipline() == MuxDiscipline.GLOBAL_ARBITRARY
			|| ( Configuration.multiplexingDiscipline() == MuxDiscipline.SERVER_LOCAL & server.useArbitraryMultiplexing() ) )
		{
			beta_loxfcaller_server = LeftOverService.arbMux( beta_loxfcaller_server, alpha_xxfcaller_server );
		}
		else
		{
			beta_loxfcaller_server = LeftOverService.fifoMux( beta_loxfcaller_server, alpha_xxfcaller_server );
		}

		if( Configuration.useGamma() == GammaFlag.GLOBALLY_ON
			|| ( Configuration.useGamma() == GammaFlag.SERVER_LOCAL & server.useGamma() ) )
		{
			alpha_xfcaller = Deconvolution.deconvolve( Convolution.convolve( alpha_xfcaller_server, gamma ), beta_loxfcaller_server );
		}
		else
		{
			alpha_xfcaller = Deconvolution.deconvolve( alpha_xfcaller_server, beta_loxfcaller_server );
		}
		
		if( Configuration.useExtraGamma() == GammaFlag.GLOBALLY_ON
			|| ( Configuration.useExtraGamma() == GammaFlag.SERVER_LOCAL & server.useExtraGamma() ) )
		{
			alpha_xfcaller = Convolution.convolve( alpha_xfcaller, gamma.removeLatency() );
		}
		
		alpha_xfcaller.beautify();
		return alpha_xfcaller;
	}

	/**
	 * Returns the output bound for a set of <code>f_xfcaller</code> from
	 * the given <code>server</code> by iteration.<br/>
	 * 
	 * @return A <code>Curve</code> instance containing the output bound
	 * @throws OutOfMemoryError
	 */
	public ArrivalCurve computeIterativeOutputBound( Server server, Set<Flow> f_xfcaller ) throws OutOfMemoryError
	{
		stack = new Stack<OutputCacheEntry>();
		internal_cache = new OutputBoundCache();

		OutputCacheEntry next = new OutputCacheEntry( server, f_xfcaller, flow_of_interest );
		stack.push( next );
		// until everything is computed
		while ( !stack.isEmpty() )
		{
			if ( !decrementOutputBoundStack() )
			{
				incrementOutputBoundStack();
			}
		}
		
		OutputCacheEntry entry = internal_cache.getEntry( server, f_xfcaller, flow_of_interest );
		
		if ( entry == null )
		{
			return ArrivalCurve.createNullArrival();
		} else {
			return entry.getArrivalCurve();	
		}
	}

	/**
	 * When the output bound for a given <code>server</code> and a set of
	 * <code>f_xfcaller</code> cannot be computed, it's because the
	 * calculation depended on results that are not yet computed, so all
	 * dependencies have to be put on the stack. That means, iterating over all
	 * the incoming links of the server.
	 */
	protected void incrementOutputBoundStack() {
		OutputCacheEntry next = stack.peek();
		
		Server server = next.getServer();
		Set<Flow> f_xfcaller = next.getBoundedFlows();
		
		Set<Flow> f_link;
		Set<Flow> f_xfcaller_link;
		Set<Flow> f_xxfcaller_link;
		
		for ( Link in_l : network.getInLinks( server ) )
		{
			f_link = network.getFlows( in_l );
			f_link.remove( flow_of_interest ); // Flow of interest has the lowest priority
			
			f_xfcaller_link = SetUtils.getIntersection( f_link, f_xfcaller );
			f_xxfcaller_link = SetUtils.getDifference( f_link, f_xfcaller_link );
			
			// Is the required output bound already computed?
			OutputCacheEntry xf_entry = internal_cache.getEntry( in_l.getSource(), f_xfcaller_link, flow_of_interest );
			
			if( xf_entry == null )
			{
				stack.push( new OutputCacheEntry( in_l.getSource(), f_xfcaller_link, flow_of_interest ) );
				// Some output bounds don't depend on anything
				if ( network.inDegree( in_l.getSource() ) == 0 || f_xfcaller_link.size() == 0 )
				{
					decrementOutputBoundStack();
				}
			}
			
			// Is the required output bound already computed?
			OutputCacheEntry xxf_entry = internal_cache.getEntry( in_l.getSource(), f_xxfcaller_link, flow_of_interest );
			
			if( xxf_entry == null )
			{
				stack.push( new OutputCacheEntry( in_l.getSource(), f_xxfcaller_link, flow_of_interest ) );
				// Some output bounds don't depend on anything
				if ( network.inDegree( in_l.getSource() ) == 0 || f_xxfcaller_link.size() == 0 )
				{
					decrementOutputBoundStack();
				}
			}
		}
	}
	
	/**
	 * This computation is nearly the same as the recursive in
	 * <code>computeRecursiveOutputBound(boolean)</code>. But the output bound
	 * is computed for the element on top of a stack. This element is a tuple
	 * with the server and the flows of interests passed as the arguments in the
	 * recursion. In many cases, the computation depends on previously computed
	 * output bounds, so when these are lacking, the calling method has to be
	 * notified. After a successful computation, the result is added to a map.
	 * 
	 * @return If the output bound of the element on top of the stack could be
	 *         computed.
	 */
	protected boolean decrementOutputBoundStack()
	{
		OutputCacheEntry next = stack.pop();
		
		Server server = next.getServer();
		Set<Flow> f_xfcaller = next.getBoundedFlows();
		
		ArrivalCurve alpha_xfcaller = ArrivalCurve.createNullArrival();
		
		if ( f_xfcaller.size() == 0 ) {
			internal_cache.addEntry( server, f_xfcaller, flow_of_interest, alpha_xfcaller );
			return true;
		}
			
		Set<Flow> f_server = network.getSourceFlows( server );
		f_server.remove( flow_of_interest ); // flow of interest has the lowest priority
		
		Set<Flow> f_xfcaller_server = SetUtils.getIntersection( f_server, f_xfcaller );
		Set<Flow> f_xxfcaller_server = SetUtils.getDifference( f_server, f_xfcaller_server );
		
		ArrivalCurve alpha_xfcaller_server = network.getSourceFlowArrivalCurve( server, f_xfcaller_server );
		ArrivalCurve alpha_xxfcaller_server = network.getSourceFlowArrivalCurve( server, f_xxfcaller_server );
		
		ArrivalCurve alpha_xfcaller_link_sum = ArrivalCurve.createNullArrival();
		ArrivalCurve alpha_xxfcaller_link_sum = ArrivalCurve.createNullArrival();
		
		Set<Flow> f_link;
		Set<Flow> f_xfcaller_link;
		Set<Flow> f_xxfcaller_link;
		
		for ( Link in_l : network.getInLinks( server ) )
		{		
			f_link = network.getFlows( in_l );
			f_link.remove( flow_of_interest ); // flow of interest has the lowest priority
			
			f_xfcaller_link = SetUtils.getIntersection( f_link, f_xfcaller );
			f_xxfcaller_link = SetUtils.getDifference( f_link, f_xfcaller_link );

			OutputCacheEntry xf_entry = internal_cache.getEntry( in_l.getSource(), f_xfcaller_link, flow_of_interest );
			if ( xf_entry == null )
			{
				stack.push( next );
				return false;
			}
			ArrivalCurve  alpha_xfcaller_link = xf_entry.getArrivalCurve();

			OutputCacheEntry xxf_entry = internal_cache.getEntry( in_l.getSource(), f_xxfcaller_link, flow_of_interest );
			if ( xxf_entry == null )
			{
				stack.push( next );
				return false;
			}
			ArrivalCurve  alpha_xxfcaller_link = xxf_entry.getArrivalCurve();

			alpha_xfcaller_link_sum = ArrivalCurve.add( alpha_xfcaller_link_sum, alpha_xfcaller_link );
			alpha_xxfcaller_link_sum = ArrivalCurve.add( alpha_xxfcaller_link_sum, alpha_xxfcaller_link );
		}

		alpha_xfcaller_server = ArrivalCurve.add( alpha_xfcaller_server, alpha_xfcaller_link_sum );
		alpha_xxfcaller_server = ArrivalCurve.add( alpha_xxfcaller_server, alpha_xxfcaller_link_sum );
		
		alpha_xfcaller_server.beautify();
		alpha_xxfcaller_server.beautify();
		
		ServiceCurve beta_loxfcaller = server.getServiceCurve();
		ServiceCurve gamma = server.getMaxServiceCurve();
		if( Configuration.multiplexingDiscipline() == MuxDiscipline.GLOBAL_ARBITRARY
			|| ( Configuration.multiplexingDiscipline() == MuxDiscipline.SERVER_LOCAL & server.useArbitraryMultiplexing() ) )
		{
			beta_loxfcaller = LeftOverService.arbMux( beta_loxfcaller, alpha_xxfcaller_server );
		}
		else
		{
			beta_loxfcaller = LeftOverService.fifoMux( beta_loxfcaller, alpha_xxfcaller_server );
		}

		if( Configuration.useGamma() == GammaFlag.GLOBALLY_ON
			|| ( Configuration.useGamma() == GammaFlag.SERVER_LOCAL & server.useGamma() ) )
		{
			alpha_xfcaller = Deconvolution.deconvolve( Convolution.convolve( alpha_xfcaller_server, gamma ), beta_loxfcaller);
		}
		else
		{
			alpha_xfcaller = Deconvolution.deconvolve( alpha_xfcaller_server, beta_loxfcaller );
		}
		
		if( Configuration.useExtraGamma() == GammaFlag.GLOBALLY_ON
			|| ( Configuration.useExtraGamma() == GammaFlag.SERVER_LOCAL & server.useExtraGamma() ) )
		{
			alpha_xfcaller = Convolution.convolve( alpha_xfcaller, gamma.removeLatency() );
		}

		internal_cache.addEntry( server, f_xfcaller, flow_of_interest, alpha_xfcaller );	
		
		return true;
	}
}

class OutputBoundCache {
	Map<Server,Set<OutputCacheEntry>> map__server__entries = new HashMap<Server,Set<OutputCacheEntry>>();
	Map<Flow,Set<OutputCacheEntry>> map__foi__entries = new HashMap<Flow,Set<OutputCacheEntry>>();
	
	// No checks for existing entry with the same variables!
	protected void addEntry( Server server, Set<Flow> flows_to_bound, Flow flow_of_interest, ArrivalCurve arrival_curve ) {
		OutputCacheEntry entry = new OutputCacheEntry( server, flows_to_bound, flow_of_interest, arrival_curve );
		
		Set<OutputCacheEntry> entries_s = map__server__entries.get( server );
		Set<OutputCacheEntry> entries_f = map__foi__entries.get( flow_of_interest );
				
		if( entries_s == null ) {
			map__server__entries.put( server, new HashSet<OutputCacheEntry>( Collections.singleton( entry ) ) );
		} else {
			entries_s.add( entry );
		}
		
		if( entries_f == null ) {
			map__foi__entries.put( flow_of_interest, new HashSet<OutputCacheEntry>( Collections.singleton( entry ) ) );
		} else {
			entries_f.add( entry );
		}
	}
	
	protected boolean addEntry( OutputCacheEntry OutputCacheEntry ) {
		// Check entry for completeness
		if( OutputCacheEntry.getServer() == null
				|| OutputCacheEntry.getFlowOfInterest() == null
				|| OutputCacheEntry.getBoundedFlows() == null
				|| OutputCacheEntry.getBoundingMethod() == null
				|| OutputCacheEntry.getMuxDiscipline() == null
				|| OutputCacheEntry.getGamma() == null
				|| OutputCacheEntry.getExtraGamma() == null
				|| OutputCacheEntry.getArrivalCurve() == null ) {
			return false;
		}
			
		
		Set<OutputCacheEntry> entries_s = map__server__entries.get( OutputCacheEntry.getServer() );
		Set<OutputCacheEntry> entries_f = map__foi__entries.get( OutputCacheEntry.getFlowOfInterest() );
				
		if( entries_s == null ) {
			map__server__entries.put( OutputCacheEntry.getServer(), new HashSet<OutputCacheEntry>( Collections.singleton( OutputCacheEntry ) ) );
		} else {
			entries_s.add( OutputCacheEntry );
		}
		
		if( entries_f == null ) {
			map__foi__entries.put( OutputCacheEntry.getFlowOfInterest(), new HashSet<OutputCacheEntry>( Collections.singleton( OutputCacheEntry ) ) );
		} else {
			entries_f.add( OutputCacheEntry );
		}
		
		return true;
	}
	
	protected OutputCacheEntry getEntry( Server server, Set<Flow> bounded_flows, Flow flow_of_interest ) {
		// Most important feature is an efficient search
		Set<OutputCacheEntry> entries_s = map__server__entries.get( server );	
		if ( entries_s == null ) {
			return null;
		}
		
		Set<OutputCacheEntry> entries_f = map__foi__entries.get( flow_of_interest );
		if ( entries_f == null ) {
			return null;
		}

		Set<OutputCacheEntry> candidates = SetUtils.getIntersection( entries_s, entries_f );
		if ( candidates == null ) {
			return null;
		}
		
		for ( OutputCacheEntry entry : candidates ) {
			if( entry.getBoundedFlows().size() == bounded_flows.size()
					&& entry.getMuxDiscipline() == Configuration.multiplexingDiscipline()
					&& entry.getGamma() == Configuration.useGamma()
					&& entry.getExtraGamma() == Configuration.useExtraGamma() ) {
				
				// should be the most expensive operation so do it last
				if( entry.getBoundedFlows().equals( bounded_flows ) ) {
					return entry;					
				}
			}
		}
		
		return null;
	}

	public void invalidateCache() {
		map__server__entries.clear();
		map__foi__entries.clear();
	}
}

class OutputCacheEntry {
	private Server server;
	private Set<Flow> bounded_flows;
	private Flow flow_of_interest;
	
	private ArrivalBoundMethods bounding_method;
	private MuxDiscipline multiplexing_discipline;
	private GammaFlag gamma;
	private GammaFlag extra_gamma;
	
	private ArrivalCurve arrival_curve;

	protected OutputCacheEntry( Server server, Set<Flow> bounded_flows, Flow flow_of_interest ) {
		this.server = server; 
		this.bounded_flows = bounded_flows;
		this.flow_of_interest = flow_of_interest;
		
		this.multiplexing_discipline = Configuration.multiplexingDiscipline();
		this.gamma = Configuration.useGamma();
		this.extra_gamma = Configuration.useExtraGamma();
		
		this.arrival_curve = null;
	}
	
	protected OutputCacheEntry( Server server, Set<Flow> bounded_flows, Flow flow_of_interest, ArrivalCurve arrival_curve ) {
		this.server = server; 
		this.bounded_flows = bounded_flows;
		this.flow_of_interest = flow_of_interest;
		
		this.multiplexing_discipline = Configuration.multiplexingDiscipline();
		this.gamma = Configuration.useGamma();
		this.extra_gamma = Configuration.useExtraGamma();
		
		this.arrival_curve = arrival_curve;
	}
	
	public Server getServer() {
		return server;
	}
	
	public Set<Flow> getBoundedFlows() {
		return bounded_flows;
	}
	
	public Flow getFlowOfInterest() {
		return flow_of_interest;
	}
	
	public ArrivalBoundMethods getBoundingMethod() {
		return bounding_method;
	}
	
	public MuxDiscipline getMuxDiscipline() {
		return multiplexing_discipline;
	}
	
	public GammaFlag getGamma() {
		return gamma;
	}
	
	public GammaFlag getExtraGamma() {
		return extra_gamma;
	}
	
	public ArrivalCurve getArrivalCurve() {
		return arrival_curve;
	}
	
	@Override
	public boolean equals( Object other ) {
		OutputCacheEntry other_entry;
		try {
			other_entry = (OutputCacheEntry) other;			
		} catch (Exception e) {
			return false;
		}
		
		if( server == other_entry.getServer()
				&& flow_of_interest == other_entry.getFlowOfInterest()
				&& bounded_flows.size() == other_entry.getBoundedFlows().size()
				&& bounding_method == other_entry.getBoundingMethod()
				&& multiplexing_discipline == other_entry.getMuxDiscipline()
				&& gamma == other_entry.getGamma()
				&& extra_gamma == other_entry.getExtraGamma()
				&& arrival_curve == other_entry.getArrivalCurve() 
				&& bounded_flows.equals( other_entry.getBoundedFlows() ) ) {			
		} else {
			return true;
		}
		return false;
	}
	
	@Override
	public String toString() {
		return "OutputCacheEntry(" + server.toString() + "; " + bounded_flows.toString() + "; "
				+ flow_of_interest.toString() + "; " + bounding_method.toString()  + "; "
				+ multiplexing_discipline.toString() + "; " + gamma.toString()  + "; "
				+ extra_gamma.toString() + "; " + arrival_curve.toString()
				+ ")";
	}
}